untyped
global function wargamesOverhaulPrecache
global function wargamesOverhaulSettings

void function wargamesOverhaulSettings(){
    #if UI
        #if S2_WG_OH_GLITCH_P_FIX
            try{
                printt("[S2.WGOH] Client has the Glitch Particle Fix ConVar set to: " + GetConVarString("glitch_oh_p_has_wg_fix"))
            }
            catch(ex){
                thread noParticleFixThread()
            }
        #else
            printt("[S2.WGOH] Client doesn't have use GlitchOverhaul, no fix required.")
        #endif


        #if S2_WG_OH_MOD_SETTINGS
            //      Post processing
            AddModTitle("S2.WargamesOverhaul")
            AddModCategory("Post-Processing")
            AddConVarSettingEnum("s2_wgoh_bloom_scale", "Bloom Scale", ["0 (None)", "0.5", "1 (Base game)", "2", "3 (High)", "4", "5 (Very High)"])
            AddConVarSettingEnum("s2_wgoh_map_fog", "Map Fog", ["Off", "On"])
            AddConVarSettingEnum("s2_wgoh_map_tint", "Sky Light Tint", ["None", "Magenta", "Cyan", "Black"])
            AddConVarSetting("s2_wgoh_cc_weight", "Colour Correction Amount", "float")
            AddConVarSettingEnum("s2_wgoh_cc_entities", "Colour Correction Strength", ["Weak", "Strong"])

            //      Base
            AddModCategory("Walls")
            AddConVarSetting("s2_wgoh_base_intensity", "Base Colour Intensity", "float")
            AddConVarSetting("s2_wgoh_wall_1_col", "Wall 1 Base Colour", "vector")
            AddConVarSetting("s2_wgoh_wall_2_col", "Wall 2 Base Colour", "vector")

            //      Kill
            AddConVarSettingEnum("s2_wgoh_kill_reactive", "Kill Reactive Walls", ["Off", "On"])
            AddConVarSetting("s2_wgoh_wall_1_kill_col", "Wall 1 Kill Colour", "vector")
            AddConVarSetting("s2_wgoh_wall_2_kill_col", "Wall 2 Kill Colour", "vector")
            AddConVarSettingEnum("s2_wgoh_kill_type", "Kill Transition Type", ["Linear", "Logarithmic", "Quadratic"])
            AddConVarSetting("s2_wgoh_kill_log_coeff", "Logarithmic Transition Coefficient", "float")
            AddConVarSetting("s2_wgoh_kill_intensity", "Kill Colour Intensity", "float")
            AddConVarSetting("s2_wgoh_kill_time", "Kill Colour Transition Time", "float")

            //      Misc
            AddConVarSettingEnum("s2_wgoh_walls_flicker", "Flicker Toggle", ["Off", "On"])
            AddConVarSetting("s2_wgoh_flicker_deviation", "Flicker Deviation", "float")
            AddConVarSetting("s2_wgoh_flicker_power", "Flicker Deviation Power", "float")
            AddConVarSettingEnum("s2_wgoh_rgb_walls", "Walls RGB Cycle", ["Off", "On"])

            //      Grid
            AddModCategory("Grid")
            AddConVarSetting("s2_wgoh_grid_colour", "Outside Grid Colour", "vector")
            AddConVarSetting("s2_wgoh_grid_intensity", "Outside Grid Intensity", "float")
            AddConVarSettingEnum("s2_wgoh_grid_rgb", "Outside Grid RGB Cycle", ["Off", "On"])
            
        #else
            thread noSettingsDiagThread()
        #endif

        #if S2_WG_OH_KILL_CALLBACK
            printt("[S2.WGOH] Client has all dependencies intstalled!")
        #else
            thread noCallbackThread()
        #endif
    #endif
}

void function wargamesOverhaulPrecache(){
    #if S2_WG_OH_KILL_CALLBACK
        #if CLIENT
            AddCallback_OnPlayerKilled(WGOH_kill)
            AddCallback_EntitiesDidLoad(WGOH_main)
            thread RegisterCC()
        #endif
    #endif
}

#if UI
void function noSettingsDiagThread(){
    WaitFrame()
    noSettingsDiag()
}

void function noSettingsDiag(){
    DialogData errorPopUp
    errorPopUp.header = "MISSING REQUIRED DEPENDENCY"
    errorPopUp.message = "WargamesOverhaul requires ModSettings to function.\nThe mod was not found; please download it."
    errorPopUp.image = $"ui/menu/common/dialog_error"
    AddDialogButton(errorPopUp, "Exit", void function(){ ClientCommand("exit") })
    AddDialogButton(errorPopUp, "Download", void function(){
        LaunchExternalWebBrowser("https://northstar.thunderstore.io/package/EladNLG/ModSettings/", WEBBROWSER_FLAG_FORCEEXTERNAL)
        noSettingsDiag()
    })
    OpenDialog(errorPopUp)
}

void function noCallbackThread(){
    WaitFrame()
    noCallbackDiag()
}

void function noCallbackDiag(){
    DialogData errorPopUp
    errorPopUp.header = "MISSING REQUIRED DEPENDENCY"
    errorPopUp.message = "WargamesOverhaul requires S2.ClientKillCallback to function.\nThe mod was not found; please download it."
    errorPopUp.image = $"ui/menu/common/dialog_error"
    AddDialogButton(errorPopUp, "Exit", void function(){ ClientCommand("exit") })
    AddDialogButton(errorPopUp, "Download", void function(){
        LaunchExternalWebBrowser("https://northstar.thunderstore.io/package/S2Mods/ClientKillCallback/", WEBBROWSER_FLAG_FORCEEXTERNAL)
        noSettingsDiag()
    })
    OpenDialog(errorPopUp)
}

void function noParticleFixThread(){
    WaitFrame()
    noParticleFixDiag()
}

void function noParticleFixDiag(){
    DialogData errorPopUp
    errorPopUp.header = "GlitchOverhaul Version Error"
    errorPopUp.message = "You're running an outdated version of GlitchOverhaul!\n\nIf you wish to use WargamesOverhaul and GlitchOverhaul at once you must update to the newest version of GlitchOverhaul."
    errorPopUp.image = $"ui/menu/common/dialog_error"
    AddDialogButton(errorPopUp, "Exit", void function(){ ClientCommand("exit") })
    AddDialogButton(errorPopUp, "Update", void function(){
        LaunchExternalWebBrowser("https://northstar.thunderstore.io/package/S2Mods/GlitchOverhaul/", WEBBROWSER_FLAG_FORCEEXTERNAL)
        noSettingsDiag()
    })
    OpenDialog(errorPopUp)
}
#endif

#if CLIENT

vector function GetConVarFloat3(string convar){
    array<string> value = split(GetConVarString(convar), " ")
    try{
        return Vector(value[0].tofloat(), value[1].tofloat(), value[2].tofloat()) 
    }
    catch(ex){
        throw "Invalid convar " + convar + "! make sure it is a float3 and formatted as \"X Y Z\""
    }
    unreachable
}

float function logAnyDecrease(float x, float coefficient){
    if(coefficient > 0.0){
        float k = (1.0/coefficient)
        float b = -(1.0/(pow(2.71828, k) - 1.0))
        float c = (1.0 + (coefficient * log(1.0/(pow(2.71828, k) - 1.0))))
        return ((coefficient * -(log(x - b))) + c)
    }
    else{
        return -1.0
    }
    unreachable
}

float function quadraticDecrease(float x){
    return ((-4 * (x - 0.5) * (x - 0.5)) + 1)
}

float function GetLogCoefficient(){
    float input = GetConVarFloat("s2_wgoh_kill_log_coeff")
    if(input > 0)
        return input
    else{
        SetConVarFloat("s2_wgoh_kill_log_coeff", 0.5)
        throw "Invalid input, Logarithmic Transition Coefficient can't be below or equal to 0!"
    }
    unreachable
}

void function SetWallColours(vector wallOneColour, vector wallTwoColour){
    SetConVarFloat("s2_wgoh_vmt_1_r", wallOneColour.x)
    SetConVarFloat("s2_wgoh_vmt_1_g", wallOneColour.y)
    SetConVarFloat("s2_wgoh_vmt_1_b", wallOneColour.z)
    SetConVarFloat("s2_wgoh_vmt_2_r", wallTwoColour.x)
    SetConVarFloat("s2_wgoh_vmt_2_g", wallTwoColour.y)
    SetConVarFloat("s2_wgoh_vmt_2_b", wallTwoColour.z)
}

void function SetGridColour(vector gridCol){
    SetConVarFloat("s2_wgoh_vmt_grid_r", gridCol.x)
    SetConVarFloat("s2_wgoh_vmt_grid_g", gridCol.y)
    SetConVarFloat("s2_wgoh_vmt_grid_b", gridCol.z)
}

var wargamesCC = null

void function RegisterCC(){
    WaitFrame()
    wargamesCC = ColorCorrection_Register("materials/correction/wargames_custom_cc.raw")
}

float killStartTime = 0.0

void function WGOH_kill(ObituaryCallbackParams inputParams){
    if(inputParams.attacker != GetLocalClientPlayer())
        return
    if(inputParams.attacker == null || !IsValid(inputParams.attacker))
        return
    killStartTime = Time()
}

const array <string> mapTints = [
    "0.0 0.0 0.0 0.0",
    "0.1 0.0 0.1",
    "0.0 0.05 0.1",
    "0.0 0.0 0.0"
]

void function WGOH_main(){
    if(GetConVarString("s2_wgoh_old_map_name") == "mp_wargames"){
        SetConVarToDefault("fog_enable")
        SetConVarToDefault("fog_enableskybox")
        SetConVarToDefault("mat_force_bloom")
        SetConVarToDefault("mat_disable_bloom")
        SetConVarToDefault("mat_bloomscale")
        SetConVarToDefault("mat_bloom_wide_amount")
        SetConVarToDefault("mat_envmap_scale")
        SetConVarToDefault("mat_sun_color")
        SetConVarToDefault("mat_sky_color")
    }
    SetConVarString("s2_wgoh_old_map_name", GetMapName())

    ColorCorrection_SetWeight(wargamesCC, 0.0)

    while(GetMapName() == "mp_wargames"){
        WaitFrame()
        if(GetMapName() != "mp_wargames")
            break
        entity player = GetLocalClientPlayer()
        if(!IsValid(player) || player == null)
            continue

        int mapFog = GetConVarInt("s2_wgoh_map_fog")
        int bloomNum = GetConVarInt("s2_wgoh_bloom_scale")
        int tintNum = GetConVarInt("s2_wgoh_map_tint")
        int ccDisableEntities = GetConVarInt("s2_wgoh_cc_entities")
        float weightCC = GetConVarFloat("s2_wgoh_cc_weight")

        float bloomScale = 0
        if(bloomNum == 1)
            bloomScale = 0.5
        if(bloomNum > 1)
            bloomScale = float(bloomNum - 1)
        
        ColorCorrection_SetWeight(wargamesCC, weightCC)

        SetConVarInt("mat_colcorrection_disableentities", ccDisableEntities)

        SetConVarInt("fog_enable", mapFog)
        SetConVarInt("fog_enableskybox", mapFog)
        SetMapSetting_FogEnabled(mapFog == 1)
        
        if(bloomNum == 0){
            SetConVarInt("mat_force_bloom", 0)
            SetConVarInt("mat_disable_bloom", 1)
        }
        else if(bloomNum == 1){
            SetConVarInt("mat_force_bloom", 1)
            SetConVarInt("mat_disable_bloom", 0)
            SetConVarFloat("mat_bloomscale", 1.0)
            SetConVarFloat("mat_bloom_wide_amount", 0.1)
        }
        else if(bloomNum == 2){
            SetConVarInt("mat_force_bloom", 1)
            SetConVarInt("mat_disable_bloom", 0)
            SetConVarToDefault("mat_bloomscale")
            SetConVarToDefault("mat_bloom_wide_amount")
        }
        else{
            SetConVarInt("mat_force_bloom", 1)
            SetConVarInt("mat_disable_bloom", 0)
            SetConVarFloat("mat_bloomscale", bloomScale)
            SetConVarFloat("mat_bloom_wide_amount", (bloomScale/7.5))
        }

        if(GetConVarString("mat_sky_color") != mapTints[tintNum])
            player.ClientCommand("mat_sky_color " + mapTints[tintNum])

        if(GetConVarString("mat_sun_color") != "0 0 0")
            player.ClientCommand("mat_sun_color 0 0 0")

        SetConVarFloat("mat_envmap_scale", 0.25)

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////        POST PROCESSING END       //////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        bool doFlicker = (GetConVarInt("s2_wgoh_walls_flicker") == 1)
        bool doKill = (GetConVarInt("s2_wgoh_kill_reactive") == 1)
        bool rgbWalls = (GetConVarInt("s2_wgoh_rgb_walls") == 1)
        bool isLinear = (GetConVarInt("s2_wgoh_kill_type") == 0)
        bool isLogarithmic = (GetConVarInt("s2_wgoh_kill_type") == 1)
        bool isQuadratic = (GetConVarInt("s2_wgoh_kill_type") == 2)
        float flickerPower = GetConVarFloat("s2_wgoh_flicker_power")
        float baseIntensity = GetConVarFloat("s2_wgoh_base_intensity")
        float killIntensity = GetConVarFloat("s2_wgoh_kill_intensity")
        float killTime = GetConVarFloat("s2_wgoh_kill_time")
        float flickerAmount = GetConVarFloat("s2_wgoh_flicker_deviation")
        float logCoefficient = GetLogCoefficient()
        vector wallOneColour = GetConVarFloat3("s2_wgoh_wall_1_col")
        vector wallTwoColour = GetConVarFloat3("s2_wgoh_wall_2_col")
        vector wallOneKillColour = GetConVarFloat3("s2_wgoh_wall_1_kill_col")
        vector wallTwoKillColour = GetConVarFloat3("s2_wgoh_wall_2_kill_col")

        float IntensityFinal = baseIntensity
        vector WallOneFinalCol = wallOneColour
        vector WallTwoFinalCol = wallTwoColour

        if(rgbWalls){
            vector rainbow
            rainbow.x = (0.5*sin(((0.5*Time()) + 0.0/3.0) * 2 * PI) + 0.5)
            rainbow.y = (0.5*sin(((0.5*Time()) + 1.0/3.0) * 2 * PI) + 0.5)
            rainbow.z = (0.5*sin(((0.5*Time()) + 2.0/3.0) * 2 * PI) + 0.5)
            WallOneFinalCol = rainbow
            WallTwoFinalCol = rainbow
        }

        if(GetConVarInt("s2_wgoh_kill_test_cvar") == 1){
            killStartTime = Time()
            SetConVarInt("s2_wgoh_kill_test_cvar", 0)
        }
        
        if(doKill && !rgbWalls){
            float elapsedTime = (Time() - killStartTime)
            if(elapsedTime <= killTime){
                float timeFrac = (elapsedTime/killTime)
                float chargeFrac = (1 - timeFrac)
                if(isLinear){
                    WallOneFinalCol = ((chargeFrac*wallOneKillColour) + (timeFrac*wallOneColour))
                    WallTwoFinalCol = ((chargeFrac*wallTwoKillColour) + (timeFrac*wallTwoColour))
                    IntensityFinal = ((killIntensity - baseIntensity) * (chargeFrac) + baseIntensity)
                }
                if(isLogarithmic){
                    float timeFracLog = logAnyDecrease(timeFrac, logCoefficient)
                    WallOneFinalCol = ((timeFracLog*wallOneKillColour) + ((1 - timeFracLog)*wallOneColour))
                    WallTwoFinalCol = ((timeFracLog*wallTwoKillColour) + ((1 - timeFracLog)*wallTwoColour))
                    IntensityFinal = ((killIntensity*timeFracLog) + ((1-timeFracLog)*baseIntensity))
                }
                if(isQuadratic){
                    float chargeFracQuad = quadraticDecrease(chargeFrac)
                    WallOneFinalCol = ((chargeFracQuad*wallOneKillColour) + ((1 - chargeFracQuad)*wallOneColour))
                    WallTwoFinalCol = ((chargeFracQuad*wallTwoKillColour) + ((1 - chargeFracQuad)*wallTwoColour))
                    IntensityFinal = (killIntensity*chargeFracQuad) + ((1-chargeFracQuad)*baseIntensity)
                }
            }
        }

        if(doFlicker){
            float upperCap = (1 + flickerAmount)
            float bottomCap = (1 - flickerAmount)
            float flickerOne = pow(RandomFloatRange(bottomCap, upperCap), flickerPower)
            float flickerTwo = pow(RandomFloatRange(bottomCap, upperCap), flickerPower)
            WallOneFinalCol = (WallOneFinalCol * flickerOne)
            WallTwoFinalCol = (WallTwoFinalCol * flickerTwo)
        }

        WallOneFinalCol = (WallOneFinalCol * IntensityFinal)
        WallTwoFinalCol = (WallTwoFinalCol * IntensityFinal)

        SetWallColours(WallOneFinalCol, WallTwoFinalCol)

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////        WALLS END       ////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        bool gridRGB = (GetConVarInt("s2_wgoh_grid_rgb") == 1)
        vector gridColour = GetConVarFloat3("s2_wgoh_grid_colour")
        float gridIntensity = GetConVarFloat("s2_wgoh_grid_intensity")

        vector GridFinalCol = gridColour

        if(gridRGB){
            GridFinalCol.x = (0.5*sin(((0.5*Time()) + 0.0/3.0) * 2 * PI) + 0.5)
            GridFinalCol.y = (0.5*sin(((0.5*Time()) + 1.0/3.0) * 2 * PI) + 0.5)
            GridFinalCol.z = (0.5*sin(((0.5*Time()) + 2.0/3.0) * 2 * PI) + 0.5)
        }

        SetGridColour(GridFinalCol * gridIntensity)

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////        GRID END       /////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    }
}

#endif